/*******************************************************************************
***  This file is part of Pecunia.                                           ***
***                                                                          ***
***  Copyright (C) 2017, 2018, 2019, 2020, 2021, 2023, 2024                  ***
***  John Schneiderman <Licencing _AT_ Schneiderman _DOT_ me>                ***
***                                                                          ***
***  This program is free software: you can redistribute it and/or modify it ***
***  under the terms of the GNU Lesser General Public License as published   ***
***  by the Free Software Foundation, either version 3 of the License, or    ***
***  (at your option) any later version.                                     ***
***                                                                          ***
***  This program is distributed in the hope that it will be useful, but     ***
***  WITHOUT ANY WARRANTY; without even the implied warranty of              ***
***  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    ***
***  See the GNU Lesser General Public License for more details.             ***
***                                                                          ***
***  You should have received a copy of the GNU Lesser General Public License***
***  along with this program. If not, see <http://www.gnu.org/licenses/>.    ***
*******************************************************************************/
#ifndef PECUNIA_EXPECTED_HPP
#define PECUNIA_EXPECTED_HPP

#include <cassert>
#include <cstddef>
#include <functional>
#include <type_traits>
#include <utility>
#include <variant>

#include <pecunia/pecunia_export.h>


namespace pecunia
{

static struct PECUNIA_EXPORT
{} inline constexpr unexpectedErr;

template<typename ErrorType>
struct Unexpected
{
	explicit Unexpected(ErrorType value);

	ErrorType value_;
};

template<typename ValueType, typename ErrorType = bool>
class Expected
{
	static_assert(
		 ! std::is_reference_v<ValueType>,
		"Cannot have an expected value type that is a reference."
	);
	static_assert(
		 ! std::is_same_v<ValueType, std::in_place_t>,
		"Cannot have an expected value type that is the std::in_place_t type."
	);
	static_assert(
		 ! std::is_same_v<ValueType, decltype(unexpectedErr)>,
		"Cannot have an expected value type that is the unexpected error type."
	);
	static_assert(
		 ! std::is_reference_v<ErrorType>,
		"Cannot have an expected error type that is a reference."
	);
	static_assert(
		 ! std::is_invocable_v<ValueType>,
		"Cannot have an expected value type that is an invocable."
	);
	static_assert(
		 ! std::is_invocable_v<ErrorType>,
		"Cannot have an expected error type that is an invocable."
	);
	static_assert(
		std::is_destructible_v<ValueType>,
		"Cannot have an expected value type that is not destructible."
	);
	static_assert(
		std::is_destructible_v<ErrorType>,
		"Cannot have an expected error type that is not destructible."
	);

	static constexpr std::size_t valueIndex{0};
	static constexpr std::size_t errorIndex{1};
	std::variant<ValueType, ErrorType> valOrErr_;

public:
	Expected();
	template<typename ... Parameters>
	Expected(std::in_place_t, Parameters&& ... params);
	template<typename ... Parameters>
	Expected(decltype(unexpectedErr), Parameters&& ... params);
	Expected(const Expected&) = default;
	Expected(Expected&&) = default;
	~Expected() = default;
	Expected& operator=(const Expected&) = default;
	Expected& operator=(Expected&&) = default;

	Expected& operator=(const ValueType& value);

	Expected& operator=(ValueType&& value);

	template<typename ErrType>
	Expected& operator=(const Unexpected<ErrType>& error);

	template<typename ErrType>
	Expected& operator=(Unexpected<ErrType>&& error);

	template<typename ErrType, typename = std::enable_if_t< ! std::is_same_v<ValueType, ErrType>>>
	Expected& operator=(const ErrType& error)
	{
		this->valOrErr_.template emplace<Expected::errorIndex>(error);
		return *this;
	}

	template<typename ErrType, typename = std::enable_if_t< ! std::is_same_v<ValueType, ErrType>>>
	Expected& operator=(ErrType&& error)
	{
		this->valOrErr_.template emplace<Expected::errorIndex>(std::move(error));
		return *this;
	}

	[[nodiscard]]
	bool hasValue() const noexcept;

	[[nodiscard]]
	const ValueType* operator->() const;

	[[nodiscard]]
	ValueType* operator->();

	[[nodiscard]]
	const ValueType& operator*() const &;

	[[nodiscard]]
	ValueType& operator*() &;

	[[nodiscard]]
	ValueType&& operator*() &&;

	[[nodiscard]]
	const ValueType& value() const &;

	[[nodiscard]]
	ValueType& value() &;

	[[nodiscard]]
	ValueType&& value() &&;

	template<typename DefaultValueType>
	[[nodiscard]]
	ValueType valueOr(DefaultValueType&& defaultVal) const;

	[[nodiscard]]
	const ErrorType& error() const &;

	[[nodiscard]]
	ErrorType& error() &;

	[[nodiscard]]
	ErrorType&& error() &&;
};

template<typename ValueType, typename ErrorType>
pecunia::Expected<ValueType, ErrorType>& applyMatch(
	pecunia::Expected<ValueType, ErrorType>& expected,
	const std::function<void(ValueType&)>& matchesValueType,
	const std::function<void(ErrorType&)>& matchesErrorType
);

template<typename ValueType, typename ErrorType>
const pecunia::Expected<ValueType, ErrorType>& applyMatch(
	const pecunia::Expected<ValueType, ErrorType>& expected,
	const std::function<void(const ValueType&)>& matchesValueType,
	const std::function<void(const ErrorType&)>& matchesErrorType
);

}

template<typename ErrorType>
pecunia::Unexpected<ErrorType>::Unexpected(ErrorType value) :
	value_{value}
{}

template<typename ValueType, typename ErrorType>
pecunia::Expected<ValueType, ErrorType>::Expected():
	valOrErr_{std::in_place_index<Expected::errorIndex>, ErrorType{}}
{}

template<typename ValueType, typename ErrorType>
template<typename ... Parameters>
pecunia::Expected<ValueType, ErrorType>::Expected(std::in_place_t, Parameters&& ... params) :
	valOrErr_{std::in_place_index<Expected::valueIndex>, std::forward<Parameters>(params)...}
{}

template<typename ValueType, typename ErrorType>
template<typename ... Parameters>
pecunia::Expected<ValueType, ErrorType>::Expected(
	decltype(unexpectedErr),
	Parameters&& ... params
) :
	valOrErr_{std::in_place_index<Expected::errorIndex>, std::forward<Parameters>(params)...}
{}

template<typename ValueType, typename ErrorType>
pecunia::Expected<ValueType, ErrorType>& pecunia::Expected<ValueType, ErrorType>::operator=(
	const ValueType& value
)
{
	this->valOrErr_.template emplace<Expected::valueIndex>(value);
	return *this;
}

template<typename ValueType, typename ErrorType>
pecunia::Expected<ValueType, ErrorType>& pecunia::Expected<ValueType, ErrorType>::operator=(
	ValueType&& value
)
{
	this->valOrErr_.template emplace<Expected::valueIndex>(std::move(value));
	return *this;
}

template<typename ValueType, typename ErrorType>
template<typename ErrType>
pecunia::Expected<ValueType, ErrorType>& pecunia::Expected<ValueType, ErrorType>::operator=(
	const pecunia::Unexpected<ErrType>& error
)
{
	this->valOrErr_.template emplace<Expected::errorIndex>(error.value_);
	return *this;
}

template<typename ValueType, typename ErrorType>
template<typename ErrType>
pecunia::Expected<ValueType, ErrorType>& pecunia::Expected<ValueType, ErrorType>::operator=(
	pecunia::Unexpected<ErrType>&& error
)
{
	this->valOrErr_.template emplace<Expected::errorIndex>(std::move(error.value_));
	return *this;
}

template<typename ValueType, typename ErrorType>
bool pecunia::Expected<ValueType, ErrorType>::hasValue() const noexcept
{
	return this->valOrErr_.index() == Expected::valueIndex;
}

template<typename ValueType, typename ErrorType>
const ValueType* pecunia::Expected<ValueType, ErrorType>::operator->() const
{
	assert(this->hasValue() && "Accessing an expected value must always have a value.");
	return &std::get<Expected::valueIndex>(this->valOrErr_);
}

template<typename ValueType, typename ErrorType>
ValueType* pecunia::Expected<ValueType, ErrorType>::operator->()
{
	assert(this->hasValue() && "Accessing an expected value must always have a value.");
	return &std::get<Expected::valueIndex>(this->valOrErr_);
}

template<typename ValueType, typename ErrorType>
const ValueType& pecunia::Expected<ValueType, ErrorType>::operator*() const &
{
	assert(this->hasValue() && "Accessing an expected value must always have a value.");
	return std::get<Expected::valueIndex>(this->valOrErr_);
}

template<typename ValueType, typename ErrorType>
ValueType& pecunia::Expected<ValueType, ErrorType>::operator*() &
{
	assert(this->hasValue() && "Accessing an expected value must always have a value.");
	return std::get<Expected::valueIndex>(this->valOrErr_);
}

template<typename ValueType, typename ErrorType>
ValueType&& pecunia::Expected<ValueType, ErrorType>::operator*() &&
{
	assert(this->hasValue() && "Accessing an expected value must always have a value.");
	return std::move(std::get<Expected::valueIndex>(this->valOrErr_));
}

template<typename ValueType, typename ErrorType>
const ValueType& pecunia::Expected<ValueType, ErrorType>::value() const &
{
	assert(this->hasValue() && "Accessing an expected value must always have a value.");
	return std::get<Expected::valueIndex>(this->valOrErr_);
}

template<typename ValueType, typename ErrorType>
ValueType& pecunia::Expected<ValueType, ErrorType>::value() &
{
	assert(this->hasValue() && "Accessing an expected value must always have a value.");
	return std::get<Expected::valueIndex>(this->valOrErr_);
}

template<typename ValueType, typename ErrorType>
ValueType&& pecunia::Expected<ValueType, ErrorType>::value() &&
{
	assert(this->hasValue() && "Accessing an expected value must always have a value.");
	return std::move(std::get<Expected::valueIndex>(this->valOrErr_));
}

template<typename ValueType, typename ErrorType>
template<typename DefaultValueType>
ValueType pecunia::Expected<ValueType, ErrorType>::valueOr(DefaultValueType&& defaultVal) const
{
	static_assert(
		std::is_convertible_v<ValueType, DefaultValueType>,
		"The default value type must be a convertable type to the expected type."
	);

	if (this->hasValue())
		return this->value();
	return defaultVal;
}

template<typename ValueType, typename ErrorType>
const ErrorType& pecunia::Expected<ValueType, ErrorType>::error() const &
{
	assert( ! this->hasValue() && "Accessing an expected error must never have a value.");
	return std::get<Expected::errorIndex>(this->valOrErr_);
}

template<typename ValueType, typename ErrorType>
ErrorType& pecunia::Expected<ValueType, ErrorType>::error() &
{
	assert( ! this->hasValue() && "Accessing an expected error must never have a value.");
	return std::get<Expected::errorIndex>(this->valOrErr_);
}

template<typename ValueType, typename ErrorType>
ErrorType&& pecunia::Expected<ValueType, ErrorType>::error() &&
{
	assert( ! this->hasValue() && "Accessing an expected error must never have a value.");
	return std::move(std::get<Expected::errorIndex>(this->valOrErr_));
}

template<typename ValueType, typename ErrorType>
pecunia::Expected<ValueType, ErrorType>& pecunia::applyMatch(
	pecunia::Expected<ValueType, ErrorType>& expected,
	const std::function<void(ValueType&)>& matchesValueType,
	const std::function<void(ErrorType&)>& matchesErrorType
)
{
	assert(matchesValueType != nullptr && "The function for a matching value cannot be null.");
	assert(matchesErrorType != nullptr && "The function for a matching error cannot be null.");

	if (expected.hasValue())
		matchesValueType(*expected);
	else
		matchesErrorType(expected.error());
	return expected;
}

template<typename ValueType, typename ErrorType>
const pecunia::Expected<ValueType, ErrorType>& pecunia::applyMatch(
	const pecunia::Expected<ValueType, ErrorType>& expected,
	const std::function<void(const ValueType&)>& matchesValueType,
	const std::function<void(const ErrorType&)>& matchesErrorType
)
{
	assert(matchesValueType != nullptr && "The function for a matching value cannot be null.");
	assert(matchesErrorType != nullptr && "The function for a matching error cannot be null.");

	if (expected.hasValue())
		matchesValueType(*expected);
	else
		matchesErrorType(expected.error());
	return expected;
}

#endif
